#include <stdio.h>
#include <stdbool.h>
#include "app_error.h"
#include "app_uart.h"
#include "nrf_delay.h"

#include "nrf_drv_saadc.h"
#include "nrf_drv_ppi.h"
#include "nrf.h"
#include "nrf_drv_timer.h"


#include "app_util_platform.h"
#include "nrf_delay.h"
#include "bsp.h"
#include "app_timer.h" 
#include "nrf_drv_spi.h"
#include "nordic_common.h"
#include "SEGGER_RTT.h"
#include "nrf_soc.h"

#include "config.h"
#include "nrf_drv_gpiote.h"

#include "softdevice_handler.h"
#include "myheader.h"

//setting up the SPI based on the spi_master_with_spi_slave example
//uses DMA
#if (SPI0_ENABLED == 1)
    static const nrf_drv_spi_t m_spi_master = NRF_DRV_SPI_INSTANCE(0);
#else
    #error "No SPI enabled."
#endif

static volatile bool m_transfer_completed = true; /**< A flag to inform about completed transfer. */

/**@brief Function for SPI master event callback.
 *
 * Upon receiving an SPI transaction complete event, checks if received data are valid.
 *
 * @param[in] spi_master_evt    SPI master driver event. 
 */
static void spi_master_event_handler(nrf_drv_spi_event_t event)
{
    uint32_t err_code = NRF_SUCCESS;
    switch (event)
    {
        //if event is equal to NRF_DRV_SPI_EVENT_DONE, then this block is executed to check if data is valid
			  case NRF_DRV_SPI_EVENT_DONE:   
            // Check if data are valid.
            err_code = bsp_indication_set(BSP_INDICATE_RCV_OK);
            APP_ERROR_CHECK(err_code);

            // Inform application that transfer is completed.
            m_transfer_completed = true;
            break;
				
				//if event is not equal to NRF_DRV_SPI_EVENT_DONE, then nothing happens
        default:
            // No implementation needed.
            break;
    }
}


/**@brief Function for initializing bsp module.
 */
void bsp_configuration()
{
		SEGGER_RTT_WriteString(0, "BSP Started\n");
    uint32_t err_code = NRF_SUCCESS;

    NRF_CLOCK->LFCLKSRC            = (CLOCK_LFCLKSRC_SRC_Xtal << CLOCK_LFCLKSRC_SRC_Pos);
    NRF_CLOCK->EVENTS_LFCLKSTARTED = 0;
    NRF_CLOCK->TASKS_LFCLKSTART    = 1;

    while (NRF_CLOCK->EVENTS_LFCLKSTARTED == 0)
    {
        // Do nothing.
    }
    APP_TIMER_INIT(APP_TIMER_PRESCALER, APP_TIMER_MAX_TIMERS, APP_TIMER_OP_QUEUE_SIZE, NULL);
    err_code = bsp_init(BSP_INIT_LED, APP_TIMER_TICKS(100, APP_TIMER_PRESCALER), NULL);
    APP_ERROR_CHECK(err_code);
}

void uart_error_handle(app_uart_evt_t * p_event)
{
    if (p_event->evt_type == APP_UART_COMMUNICATION_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_communication);
    }
    else if (p_event->evt_type == APP_UART_FIFO_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_code);
    }
}

void uart_config(void){
		uint32_t err_code;
		const app_uart_comm_params_t comm_params =
      {
          RX_PIN_NUMBER,
          TX_PIN_NUMBER,
          RTS_PIN_NUMBER,
          CTS_PIN_NUMBER,
          APP_UART_FLOW_CONTROL_ENABLED,
          false,
          UART_BAUDRATE_BAUDRATE_Baud38400
      };

    APP_UART_FIFO_INIT(&comm_params,
                         UART_RX_BUF_SIZE,
                         UART_TX_BUF_SIZE,
                         uart_error_handle,
                         APP_IRQ_PRIORITY_LOW,
                         err_code);
	
		APP_ERROR_CHECK(err_code);
}

/**
 * @brief Work-around for transmitting 1 byte with SPIM.
 *
 * @param spim: The SPIM instance that is in use.
 * @param ppi_channel: An unused PPI channel that will be used by the
		workaround.
 * @param gpiote_channel: An unused GPIOTE channel that will be used by
		the workaround.
 *
 * @warning Must not be used when transmitting multiple bytes.
 * @warning After this workaround is used, the user must reset the PPI
		channel and the GPIOTE channel before attempting to transmit multiple
		bytes.
 */
void setup_workaround_for_ftpan_58(NRF_SPIM_Type *spim, 
																	uint32_t ppi_channel, 
																	uint32_t gpiote_channel)
{
	// Create an event when SCK toggles.
	NRF_GPIOTE->CONFIG[gpiote_channel] = 
		(GPIOTE_CONFIG_MODE_Event << GPIOTE_CONFIG_MODE_Pos) | 
		(spim->PSEL.SCK << GPIOTE_CONFIG_PSEL_Pos) | 
		(GPIOTE_CONFIG_POLARITY_Toggle <<GPIOTE_CONFIG_POLARITY_Pos);

	// Stop the spim instance when SCK toggles.
	NRF_PPI->CH[ppi_channel].EEP = (uint32_t)&NRF_GPIOTE->EVENTS_IN[gpiote_channel];
	NRF_PPI->CH[ppi_channel].TEP = (uint32_t)&spim->TASKS_STOP;
	NRF_PPI->CHENSET = 1U << ppi_channel;
}

/**@brief Functions prepares buffers and starts data transfer to AS3911
 *
 * @param[in] p_tx_data     A pointer to a buffer TX.
 * @param[in] p_rx_data     A pointer to a buffer RX.
 * @param[in] len           A length of the data buffers.
 */
static void spi_send_recv(uint8_t * const p_tx_data,
                          uint8_t * const p_rx_data,
                          const uint16_t  len)
{
		SEGGER_RTT_WriteString(0, "Send and Receive info \n");
		uint32_t err_code;
	
    if (len == 1){
			nrf_ppi_channel_t ppi_channel;
			uint32_t gpiote_task_addr;
			ret_code_t err_code;
			
			//initializes the GPIOTE channel so that SCK toggles
			nrf_drv_gpiote_out_config_t config = GPIOTE_CONFIG_OUT_TASK_TOGGLE(false);
			err_code = nrf_drv_gpiote_out_init(SPIM0_SCK_PIN, &config);
			APP_ERROR_CHECK(err_code);	
	
			//allocates the first unused PPI Channel
			err_code = nrf_drv_ppi_channel_alloc(&ppi_channel);
			APP_ERROR_CHECK(err_code);
	
			//get the address of the GPIOTE task 
			//(is this is a GPIOTE channel??????)
			gpiote_task_addr = nrf_drv_gpiote_out_task_addr_get	(SPIM0_SCK_PIN);
	
			//enable the PPI channel.
			err_code = nrf_drv_ppi_channel_enable(ppi_channel);
			APP_ERROR_CHECK(err_code);
	
			//enable the GPIOTE output pin task.
			nrf_drv_gpiote_out_task_enable(SPIM0_SCK_PIN);
			
			//Start transfer of data
			err_code = nrf_drv_spi_transfer(&m_spi_master, p_tx_data, len, p_rx_data, len);
			APP_ERROR_CHECK(err_code);
			
			//interrupt transfer of data so that only one byte is sent
			setup_workaround_for_ftpan_58(NRF_SPIM_Type *spim, 
																ppi_channel, 
																gpiote_task_addr);
																
			//disable the ppi channel
			err_code = nrf_drv_ppi_channel_disable(ppi_channel);
	
			//disable a GPIOTE output pin task.
			nrf_drv_gpiote_out_task_disable(SPIM0_SCK_PIN);
	
			//uninitializes the gpiote channel
			//the gpiote channel is represented by a 32-bit variable
			//not sure if we need to unintitialize the pin, or if we can just disable it
			//nrf_drv_gpiote_out_uninit(SPIM0_SCK_PIN);	
		}
		else{
			// Start transfer.
			err_code = nrf_drv_spi_transfer(&m_spi_master, p_tx_data, len, p_rx_data, len);
			APP_ERROR_CHECK(err_code);
		}
		
    nrf_delay_ms(delay);
}


/**@brief Function for sending two bytes to AS3911
*
* @param[in] reg is the register of the AS3911 that you want to access
* @param[in] val is the data written to the AS3911 through SPI
*/
void as3911WriteRegister(uint8_t reg, uint8_t val){
	uint8_t write_buffer[2];
	SEGGER_RTT_WriteString(0, "Entered Write Register\n");
	write_buffer[0] = reg | AS3911_WRITE_MODE;
	write_buffer[1] = val;
	spi_send_recv(write_buffer, NULL, 2);
}

/**@brief Function for sending 1 byte to AS3911
*
* @param[in] cmd is the register of the AS3911 that you want to write to
*/
void as3911ExecuteCommand(uint8_t cmd){
	uint8_t write_buffer[1];
	SEGGER_RTT_WriteString(0, "Entered Execute Command\n");
	write_buffer[0] = cmd | AS3911_CMD_MODE;
	spi_send_recv(write_buffer, NULL, 1);
}

int main(void){
		//0 is the terminal number where the text is output on the J-link RTT
		//These lines are meant for debugging purposes and are not necessary 
		SEGGER_RTT_WriteString(0, "Program Begin\n");			
		
		//INITIALIZING PERIPHERALS
		//Initialize UART, and board support
		uart_config();
		bsp_configuration();
		
		//Configure SPI interface
    nrf_drv_spi_config_t const config =
    {
        #if (SPI0_ENABLED == 1)
            .sck_pin  = SPIM0_SCK_PIN, 
            .mosi_pin = SPIM0_MOSI_PIN,
            .miso_pin = SPIM0_MISO_PIN,
            .ss_pin   = SPIM0_SS_PIN,
        
        #endif
        .irq_priority = APP_IRQ_PRIORITY_LOW, //tells computer what to prioritize over other things
				.orc          = 0x00,
        .frequency    = NRF_DRV_SPI_FREQ_1M,
        .mode         = NRF_DRV_SPI_MODE_1,
        .bit_order    = NRF_DRV_SPI_BIT_ORDER_MSB_FIRST,
				
    };
		
		//Initialize SPI
		ret_code_t err_code = nrf_drv_spi_init(&m_spi_master, &config, spi_master_event_handler);
    APP_ERROR_CHECK(err_code);
		
		//initialize GPIOTE and PPI
    err_code = nrf_drv_ppi_init();
    APP_ERROR_CHECK(err_code);

    err_code = nrf_drv_gpiote_init();
    APP_ERROR_CHECK(err_code);
		
		//VARIABLES
		uint8_t numBytesReg = 0x1D;
		
		//Infinite loop containing main functionality
		for(;;){
			if (m_transfer_completed)
			{
				m_transfer_completed = false;
		
				as3911WriteRegister(numBytesReg, 0x01);
				as3911ExecuteCommand(0xC1);
			}	
		}
}		
